"""
pyRFC3339 parses and generates :RFC:`3339`-compliant timestamps using Python
:class:`datetime.datetime` objects.

>>> from pyrfc3339 import generate, parse
>>> from datetime import datetime, timezone
>>> generate(datetime.now(timezone.utc)) #doctest:+ELLIPSIS
'...T...Z'
>>> parse('2009-01-01T10:01:02Z')
datetime.datetime(2009, 1, 1, 10, 1, 2, tzinfo=datetime.timezone.utc)
>>> parse('2009-01-01T14:01:02-04:00')
datetime.datetime(2009, 1, 1, 14, 1, 2, tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=72000)))

"""

from importlib.metadata import PackageNotFoundError, version

from .generator import generate
from .parser import parse

try:
    __version__ = version("pyrfc3339")
except PackageNotFoundError:
    pass

__all__ = ["generate", "parse"]



import re
import sys
from datetime import datetime, timezone

from .utils import datetime_utcoffset


def parse(timestamp: str, utc: bool = False, produce_naive: bool = False) -> datetime:
    """
    Parse an :RFC:`3339`-formatted timestamp and return a :class:`datetime.datetime`.

    If the timestamp is presented in UTC, then the :attr:`~datetime.datetime.tzinfo` attribute of the
    returned `datetime` will be set to :attr:`datetime.timezone.utc`.

    >>> parse('2009-01-01T10:01:02Z')
    datetime.datetime(2009, 1, 1, 10, 1, 2, tzinfo=datetime.timezone.utc)

    Otherwise, a :class:`datetime.timezone` instance is created with the appropriate offset, and
    the :attr:`~datetime.datetime.tzinfo` attribute of the returned :class:`~datetime.datetime` is set to that value.

    >>> parse('2009-01-01T14:01:02-04:00')
    datetime.datetime(2009, 1, 1, 14, 1, 2, tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=72000)))

    However, if :meth:`parse()`  is called with :python:`utc=True`, then the returned
    :class:`~datetime.datetime` will be normalized to UTC (and its :attr:`~datetime.datetime.tzinfo` attribute set to
    :attr:`~datetime.timezone.utc`), regardless of the input timezone.

    >>> parse('2009-01-01T06:01:02-04:00', utc=True)
    datetime.datetime(2009, 1, 1, 10, 1, 2, tzinfo=datetime.timezone.utc)

    As parsing is delegated to :meth:`datetime.datetime.fromisoformat()`, certain
    timestamps which do not strictly adhere to :RFC:`3339` are nonetheless accepted.

    >>> parse('2009-01-01T06:01:02')
    datetime.datetime(2009, 1, 1, 6, 1, 2)

    Exceptions will, however, be thrown for blatantly invalid input:

    >>> parse('2009-01-01T25:01:02Z') # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    ValueError: hour must be in 0..23

    :param str timestamp: the :RFC:`3339` timestamp to be parsed
    :param bool utc: :const:`True` to normalize the timestamp to UTC; :const:`False` otherwise. Defaults to :const:`False`.
    :param bool produce_naive: :const:`True` if the produced :class:`~datetime.datetime` instance should
                               not have a timezone attached (that is, be 'naive'); :const:`False` otherwise.
                               Defaults to :const:`False`.
    :return: the parsed timestamp
    :rtype: datetime.datetime

    """

    # Python does not recognize "Z" as an alias for "+00:00", so we perform the
    # substitution here.
    timestamp = re.sub("Z$", "+00:00", timestamp, flags=re.IGNORECASE)

    # Python releases prior to 3.11 only support three or six digits of fractional
    # seconds. RFC 3339 is more lenient, so pad to six digits and truncate any
    # excessive digits.
    # This can be removed in October 2026, once Python 3.10 and earlier
    # have been retired.
    # noinspection PyUnreachableCode
    if sys.version_info < (3, 11):
        timestamp = re.sub(
            r"(\.)([0-9]+)(?=[+\-][0-9]{2}:[0-9]{2}$)",
            lambda match: match.group(1) + match.group(2).ljust(6, "0")[:6],
            timestamp,
        )

    dt_out = datetime.fromisoformat(timestamp)

    if utc:
        dt_out = dt_out.astimezone(timezone.utc)

    if produce_naive:
        if datetime_utcoffset(dt_out) == 0:
            dt_out = dt_out.replace(tzinfo=None)
        else:
            raise ValueError("cannot produce a naive datetime from a local timestamp")

    return dt_out



from datetime import datetime


def datetime_utcoffset(dt: datetime) -> float:
    """
    Return the UTC offset for an aware :class:`datetime.datetime` in seconds.

    >>> from datetime import datetime
    >>> from zoneinfo import ZoneInfo
    >>> z = ZoneInfo('US/Eastern')
    >>> dt = datetime(2024, 11, 5, 19, 7, 6, tzinfo=z)
    >>> datetime_utcoffset(dt)
    -18000.0

    >>> dt = datetime(2024, 11, 5, 19, 7, 6)
    >>> datetime_utcoffset(dt)
    Traceback (most recent call last):
    ...
    AssertionError

    :param datetime.datetime dt: a :class:`~datetime.datetime` instance; must be aware (that is, have a timezone attached)
    :return: the UTC offset of the supplied :class:`~datetime.datetime` in seconds
    :rtype: float

    """

    assert dt.tzinfo is not None

    tz = dt.tzinfo
    offset = tz.utcoffset(dt)

    assert offset is not None

    return offset.total_seconds()



import re
from datetime import datetime, timezone


def generate(
    dt: datetime,
    utc: bool = True,
    accept_naive: bool = False,
    microseconds: bool = False,
) -> str:
    """
    Generate an :RFC:`3339`-formatted timestamp from a :class:`datetime.datetime`.

    >>> from datetime import datetime, timezone
    >>> from zoneinfo import ZoneInfo
    >>> generate(datetime(2009, 1, 1, 12, 59, 59, 0, timezone.utc))
    '2009-01-01T12:59:59Z'

    The timestamp be normalized to UTC unless :python:`utc=False` is specified, in which case
    it will use the timezone from the :class:`~datetime.datetime`'s :attr:`~datetime.datetime.tzinfo` attribute.

    >>> eastern = ZoneInfo('US/Eastern')
    >>> dt = datetime(2009, 1, 1, 12, 59, 59, tzinfo=eastern)
    >>> generate(dt)
    '2009-01-01T17:59:59Z'
    >>> generate(dt, utc=False)
    '2009-01-01T12:59:59-05:00'

    Unless :python:`accept_naive=True` is specified, the :class:`~datetime.datetime` must not be naive.

    >>> generate(datetime(2009, 1, 1, 12, 59, 59, 0))
    Traceback (most recent call last):
    ...
    ValueError: naive datetime and accept_naive is False

    >>> generate(datetime(2009, 1, 1, 12, 59, 59, 0), accept_naive=True)
    '2009-01-01T12:59:59Z'

    If, however, :python:`accept_naive=True` is specified, the :class:`~datetime.datetime` is assumed to represent a UTC time.
    Attempting to generate a local timestamp from a naive datetime will result in an error.

    >>> generate(datetime(2009, 1, 1, 12, 59, 59, 0), accept_naive=True, utc=False)
    Traceback (most recent call last):
    ...
    ValueError: cannot generate a local timestamp from a naive datetime

    :param datetime.datetime dt: the :class:`~datetime.datetime` for which to generate an :RFC:`3339` timestamp.
    :param bool utc: :const:`True` to normalize the supplied :class:`datetime.datetime` to UTC; :const:`False` otherwise.
                     Defaults to :const:`True`.
    :param bool accept_naive: :const:`True` if :func:`generate()` should accept a 'naive' datetime
                              (that is, one without timezone information) and treat it as a UTC timestamp;
                              :const:`False` otherwise. Defaults to :const:`False`.
    :param bool microseconds: :const:`True` to generate a timestamp which includes fractional seconds, if present;
                              :const:`False` otherwise. Defaults to :const:`False`.
                              Note that fractional seconds are *truncated*,
                              not rounded when :obj:`microseconds` is :const:`False`.
    :return: the supplied :class:`~datetime.datetime` instance represented as an :RFC:`3339` timestamp
    :rtype: str

    """

    if dt.tzinfo is None:
        if accept_naive:
            if utc:
                dt = dt.replace(tzinfo=timezone.utc)
            else:
                raise ValueError(
                    "cannot generate a local timestamp from a naive datetime"
                )
        else:
            raise ValueError("naive datetime and accept_naive is False")

    if utc:
        dt = dt.astimezone(timezone.utc)

    timestamp = dt.isoformat(timespec="microseconds" if microseconds else "seconds")

    if dt.tzinfo == timezone.utc:
        timestamp = re.sub(r"\+00:00$", "Z", timestamp)

    return timestamp
